open Unix
open Sqlite
open Sqlite_util

let string_after s n =
  String.sub s n (String.length s - n)

let logfile = ref Pervasives.stdout

let open_log file =
   logfile := open_out_gen [Open_append;Open_creat;Open_text] 0o644 file

let log s = 
   let curtime = Strftime.strftime ~tm:(localtime (time ())) "%d/%m %H:%M " in
      output_string !logfile (curtime ^ s);
      flush !logfile

let open_db file =
   let db = Sqlite.db_open file in
      if not (result_bool db
         "SELECT name FROM SQLITE_MASTER WHERE type='table' AND name='stamps'")
      then begin try
         exec db 
	  "CREATE TABLE stamps (triple varchar, stamp float, counter integer )";
         exec db "create index stamps_idx on stamps(triple)";
      with Sqlite_error s -> 
         raise (Failure "error while creating table")
      end;
      db

let check_db db triple new_time =
   let tr = escape triple in
   let vm = compile_simple db
      ("SELECT stamp, counter FROM stamps WHERE triple=" ^ tr) in
      try
	 let result = step_simple vm in
	    finalize vm;
	    if int_of_string result.(1) < 3 then
	       exec db ("UPDATE stamps SET counter=counter+1 WHERE triple=" ^
			   tr);
	    float_of_string result.(0), int_of_string result.(1)
      with Sqlite_done ->
	 exec db ("INSERT INTO stamps " ^ 
		     values [escape triple; 
			     escape (string_of_float new_time);
			     "0"]);
	 new_time, 0
       
let cleanup_db db () =
   let now = string_of_float (gettimeofday () -. 15000.) in
      exec db ("DELETE FROM stamps WHERE counter=0 AND substr(stamp, 0, 10) < "
	       ^ String.sub now 0 10)

let rec get_attrs in_chan acc =
   let line = input_line in_chan in
      if line = "" || line.[0] = '\n' || line.[0] = '\r' then 
	 acc
      else
	 let eq = String.index line '=' in
	    get_attrs in_chan 
	       ((String.sub line 0 eq, string_after line (eq+1)) :: acc)

let find_attr attr attrs =
   try List.assoc attr attrs with Not_found -> ""

let smtpd_access_policy attrs db grey_delay my_helo my_recipients = 
   let helo = find_attr "helo_name" attrs in
   let recipient = String.lowercase (find_attr "recipient" attrs) in
   let sender = String.lowercase (find_attr "sender" attrs) in
   let client_address = find_attr "client_address" attrs in
      if List.mem helo my_helo then begin
	 log (Printf.sprintf "action=rekect: HELO %s from %s %s\n"
		 helo sender client_address);
	 "reject"
      end
      else if not (List.mem recipient my_recipients) then begin
	 log (Printf.sprintf "action=reject: recipient %s from %s %s\n"
		 recipient sender client_address);
	 "reject"
      end
      else begin
	 let triple = client_address ^ "/" ^ sender ^ "/" ^ recipient in
	 let now = time () in
	 let time_stamp, counter = check_db db triple now in
	    if counter = 3 then begin
	       log (Printf.sprintf "action=OK: %s %s -> %s\n"
		       sender client_address recipient);
	       "OK"
	    end
	    else if now -. time_stamp > grey_delay then begin
	       log (Printf.sprintf "action=dunno: %s %s -> %s\n"
		       sender client_address recipient);
	       "dunno"
	    end
	    else begin
	       log (Printf.sprintf "action=defer_if_permit: %s %s -> %s\n"
		       sender client_address recipient);
	       "defer_if_permit Service is unavailable"
	    end
      end

let rec policy db grey_delay my_helo my_recipients in_chan out_chan =
   let attrs = get_attrs in_chan [] in
      List.iter (fun (a,b) ->
		    log (Printf.sprintf "%s = %s\n" a b)) attrs;
      log "\n";
      begin
	 match find_attr "request" attrs with
	    | "smtpd_access_policy" ->
		 let reply = smtpd_access_policy attrs 
		    db grey_delay my_helo my_recipients in
		    output_string out_chan 
		       (Printf.sprintf "action=%s\n\n" reply);
		    flush out_chan;
	    | _ -> 
		 log (Printf.sprintf "Other request: %s\n"
			 (find_attr "request" attrs));
		 output_string out_chan "\n"
      end;
      log "\n";
      policy db grey_delay my_helo my_recipients in_chan out_chan

let init () =
   let cfg = Config.parse "greybottle.conf" in
   let lf = (List.hd (List.assoc "logfile" cfg)) in
      if !logfile <> Pervasives.stdout then
	 close_out !logfile;
      open_log lf;
      
      let grey_delay = 
	 float_of_string (List.hd (List.assoc "grey_delay" cfg)) in
      let my_helo = List.assoc "my_helo" cfg in
      let my_recipients = List.assoc "my_recipients" cfg in
      let db = open_db (List.hd (List.assoc "db_file" cfg)) in
      let port = int_of_string (List.hd (List.assoc "port" cfg)) in
	 db, port, grey_delay, my_helo, my_recipients, lf

let _ =
   if Array.length Sys.argv > 1 && Sys.argv.(1) = "-v" then begin
      print_endline 
	 "Copyright (c) 2005, Anastasia Gornostaeva, ermine@ermine.pp.ru";
      print_endline "Grey Bottle version 0.1"
   end
   else begin
      let db, port, grey_delay, my_helo, my_recipients, lf = init () in
      let inet_addr = inet_addr_any in
      let _rid = Scheduler.add_task (cleanup_db db) 
	 (gettimeofday () +. 7200.) (fun () -> 86400.) in
      let () =
	 Sys.set_signal Sys.sigusr1
            (Sys.Signal_handle (fun x -> 
				   close_out !logfile;
				   open_log lf)) in

				   
	 log "Starting...\n";
	 let sockaddr = ADDR_INET (inet_addr, port) in
	    try
	       establish_server (policy db grey_delay my_helo my_recipients) 
		  sockaddr;
	    with
	       | Unix_error (code, _, _) ->
		    log (Printf.sprintf "%s\n" (error_message code))
	       | End_of_file ->
		    log "Postfix closed a connection\n"
	       | exn ->
		    log (Printf.sprintf "Exception: %s\n" 
			    (Printexc.to_string exn))
   end
